use std::{collections::HashMap, error::Error};

use helmzoo_lib::{
    cli::is_path_relative,
    config::ConfigImpl,
    helm::{chart::Chart, repository::Repository},
    include::Include,
    output::message_info,
};
use serde::{Deserialize, Serialize};

use crate::{extensions::Extension, mirror::Mirror, patches::Patch};

#[derive(Serialize, Deserialize, PartialEq, Debug, Clone, PartialOrd, Ord, Eq)]
pub(crate) enum SupportedIncludes {
    Repositories,
    Mirrors,
    Charts,
}

#[derive(Serialize, Deserialize, PartialEq, Debug, Clone, Default)]
pub(crate) struct ChartExtended {
    #[serde(flatten)]
    pub(crate) chart: Chart,
    pub(crate) extensions: Option<Vec<Extension>>,
    pub(crate) patches: Option<Vec<Patch>>,
    pub variables: Option<HashMap<String, String>>,
}

impl ChartExtended {
    pub(crate) fn populate_variables(&mut self, global_variables: Option<HashMap<String, String>>) {
        if let Some(global_vars) = global_variables {
            self.variables = match self.variables.clone() {
                Some(mut vars) => {
                    vars.extend(global_vars);
                    Some(vars)
                }
                None => Some(global_vars),
            }
        };
    }
}

#[derive(Serialize, Deserialize, PartialEq, Debug, Clone, Default)]
pub(crate) struct Config {
    pub(crate) variables: Option<HashMap<String, String>>,
    #[serde(default = "empty_vec")]
    pub(crate) repositories: Vec<Repository>,
    pub(crate) include: Option<Vec<Include<SupportedIncludes>>>,
    #[serde(default = "empty_vec")]
    pub(crate) charts: Vec<ChartExtended>,
    pub(crate) patches: Option<Vec<Patch>>,
    #[serde(default = "empty_vec")]
    pub(crate) mirrors: Vec<Mirror>,
}

fn empty_vec<T>() -> Vec<T> {
    vec![]
}

impl ConfigImpl for Config {
    fn apply_includes(&mut self, config_path: String) -> Result<(), Box<dyn Error>> {
        if let Some(mut include) = self.include.clone() {
            include.sort_by_key(|f| f.kind.clone());
            include
                .iter()
                .try_for_each(|i| -> Result<(), Box<dyn Error>> {
                    let include_path = match is_path_relative(i.path.clone()) {
                        true => format!("{}/{}", config_path, i.path),
                        false => i.path.clone(),
                    };
                    match i.kind {
                        SupportedIncludes::Charts => {
                            Ok(self.append_to_charts(include_charts(include_path)?))
                        }
                        SupportedIncludes::Repositories => {
                            Ok(self.append_to_repositories(include_repositories(include_path)?))
                        }
                        SupportedIncludes::Mirrors => todo!(),
                    }
                })?
        }
        Ok(())
    }
}

impl Config {
    // TODO: Maybe it can be a generic function
    fn append_to_repositories(&mut self, mut repositories: Vec<Repository>) {
        self.repositories.append(&mut repositories);
    }

    fn append_to_charts(&mut self, mut charts: Vec<ChartExtended>) {
        self.charts.append(&mut charts);
    }
}

fn include_repositories(path: String) -> Result<Vec<Repository>, Box<dyn Error>> {
    message_info(&format!(
        "trying to include repositories from {}",
        path.clone()
    ));
    let file = std::fs::File::open(path.clone())?;
    let repositories: Vec<Repository> = match serde_yaml::from_reader(file) {
        Ok(res) => res,
        Err(_) => {
            let file = std::fs::File::open(path.clone())?;
            let repo: Repository = serde_yaml::from_reader(file)?;
            vec![repo]
        }
    };
    Ok(repositories)
}

fn include_charts(path: String) -> Result<Vec<ChartExtended>, Box<dyn Error>> {
    message_info(&format!("trying to include chart from {}", path.clone()));
    let file = std::fs::File::open(path.clone())?;

    let chart_dir = match std::path::Path::new(&path).parent() {
        Some(dir) => match dir.to_str() {
            Some(dir) => dir.to_string(),
            None => {
                return Err(Box::from(format!(
                    "chart parrent dir not found for {}",
                    path
                )));
            }
        },
        None => {
            return Err(Box::from(format!(
                "chart parrent dir not found for {}",
                path
            )));
        }
    };

    let mut charts: Vec<ChartExtended> = match serde_yaml::from_reader(file) {
        Ok(res) => res,
        Err(_) => {
            let file = std::fs::File::open(path.clone())?;
            let chart: ChartExtended = serde_yaml::from_reader(file)?;
            vec![chart]
        }
    };

    charts.iter_mut().for_each(|chart| {
        match chart.extensions {
            Some(ref mut extensions) => extensions.iter_mut().for_each(|extension| {
                if is_path_relative(extension.source_dir.clone()) {
                    let clean_path = match extension.source_dir.clone().starts_with("./") {
                        true => extension.source_dir.clone().replacen("./", "", 1),
                        false => extension.source_dir.clone(),
                    };
                    if is_path_relative(clean_path.clone()) {
                        let new_path = format!("{}/{}", chart_dir, clean_path);
                        extension.source_dir = new_path;
                    }
                }
            }),
            None => message_info(&format!("no extensions set, nothing to update")),
        };
        match chart.patches {
            Some(ref mut patches) => patches.iter_mut().for_each(|patch| {
                if is_path_relative(patch.get_path().clone()) {
                    let clean_path = match patch.get_path().clone().starts_with("./") {
                        true => patch.get_path().clone().replacen("./", "", 1),
                        false => patch.get_path().clone(),
                    };
                    if is_path_relative(clean_path.clone()) {
                        let new_path = format!("{}/{}", chart_dir, clean_path);
                        patch.set_path(new_path);
                    }
                }
            }),
            None => message_info(&format!("no patch set, nothing to update")),
        };
    });
    Ok(charts)
}

#[cfg(test)]
mod tests {
    use std::{error::Error, fs::File, io::Write};

    use helmzoo_lib::{
        config::ConfigImpl,
        helm::{helm_repository::HelmRepo, repository::Repository},
        include::Include,
    };
    use tempfile::tempdir;

    use crate::config::SupportedIncludes;

    use super::{include_repositories, Config};

    fn prepare_test_file(name: &str, data: &str) -> Result<String, Box<dyn Error>> {
        let dir = tempdir()?;
        let file_path = dir.into_path().join(&name);
        let mut file = File::create(file_path.clone())?;
        file.write_all(data.as_bytes())?;
        let path = file_path.into_os_string().to_str().unwrap().to_string();
        Ok(path)
    }

    #[test]
    fn test_cfg_apply_includes_repo() -> Result<(), Box<dyn Error>> {
        let data = "---
- name: test
  helm:
    url: test.rocks
            ";
        let path = prepare_test_file("repositories.yaml", data)?;
        let repo_1: Repository = Repository {
            name: "test".to_string(),
            helm: None,
            git: None,
        };
        let repos = vec![repo_1.clone()];
        let includes: Vec<Include<SupportedIncludes>> = vec![Include {
            path: path.clone(),
            kind: SupportedIncludes::Repositories,
        }];
        let mut config = Config {
            repositories: repos.clone(),
            include: Some(includes),
            charts: vec![],
            variables: None,
            patches: None,
            mirrors: vec![],
        };
        let repo_2 = Repository {
            name: "test".to_string(),
            helm: Some(HelmRepo {
                url: "test.rocks".to_string(),
            }),
            git: None,
        };
        config.apply_includes(path)?;
        assert_eq!(config.repositories, vec!(repo_1, repo_2));
        Ok(())
    }
    #[test]
    fn test_cfg_append_to_repos() -> Result<(), Box<dyn Error>> {
        let repo_1: Repository = Repository {
            name: "test".to_string(),
            helm: None,
            git: None,
        };
        let repos = vec![repo_1.clone()];
        let mut config = Config {
            variables: None,
            repositories: repos.clone(),
            include: None,
            mirrors: vec![],
            charts: vec![],
            patches: None,
        };
        let repo_2 = Repository {
            name: "test_2".to_string(),
            helm: None,
            git: None,
        };
        let repo_2_vec = vec![repo_2.clone()];
        config.append_to_repositories(repo_2_vec);
        assert_eq!(config.repositories, vec!(repo_1, repo_2));
        Ok(())
    }

    #[test]
    fn test_include_repositories() -> Result<(), Box<dyn Error>> {
        let data = "---
- name: test
  helm:
    url: test.rocks
            ";
        let path = prepare_test_file("repositories.yaml", data)?;
        let include = include_repositories(path)?;
        let expected: Vec<Repository> = vec![Repository {
            name: "test".to_string(),
            git: None,
            helm: Some(HelmRepo {
                url: "test.rocks".to_string(),
            }),
        }];
        assert_eq!(expected, include);
        Ok(())
    }

    #[test]
    fn test_include_repository() -> Result<(), Box<dyn Error>> {
        let data = "---
name: test
helm:
  url: test.rocks
            ";
        let path = prepare_test_file("repositories.yaml", data)?;
        let include = include_repositories(path)?;
        let expected: Vec<Repository> = vec![Repository {
            name: "test".to_string(),
            git: None,
            helm: Some(HelmRepo {
                url: "test.rocks".to_string(),
            }),
        }];
        assert_eq!(expected, include);
        Ok(())
    }
}
