use clap::Parser;
use config::Config;
use helmzoo_lib::{
    self,
    cli::{check_prerequisites, get_full_path_dir},
    config::{read_config, ConfigImpl},
    helm::repository::RepositoryImpl,
    output::{message_empty, message_error},
};
use std::{error::Error, process::exit};

use crate::mirror::mirror_from_mirror_obj;

mod config;
mod extensions;
mod mirror;
mod patches;

/// Simple program to greet a person
#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
    /// Name of the working dir
    #[arg(short, long)]
    workdir: Option<String>,
    /// Path to the configuration file
    #[arg(short, long)]
    config: String,
    /// Dry run
    #[arg(short, long, default_value = "false")]
    dry_run: bool,
    #[arg(long, default_value = "false")]
    skip_prerequisites_check: bool,
    /// Init git patch. Use it if you want to create git patch for a chart
    /// It's going to pull a chart and init a git repo there, so you can
    /// apply changes and create a patch file
    /// It's not going to try mirroring changes, but will apply extensions
    /// and patches that are already defined
    #[arg(long)]
    init_git_patch: Option<Vec<String>>,
    #[arg(long, default_value = "helm")]
    helm_bin: String,
    #[arg(long, default_value = "git")]
    git_bin: String,
    #[arg(long, default_value = "yq")]
    yq_bin: String,
}

fn exec(args: Args) -> Result<(), Box<dyn Error>> {
    let prerequisites = vec![args.helm_bin, args.git_bin.clone(), args.yq_bin];
    check_prerequisites(prerequisites)?;
    let workdir_path = helmzoo_lib::workdir::setup_workdir(args.workdir)?;
    let mut config: Config = read_config(args.config.clone())?;
    let config_full_path = get_full_path_dir(args.config.clone())?;
    config.apply_includes(config_full_path.clone())?;
    config
        .charts
        .into_iter()
        .try_for_each(|mut chart| -> Result<(), Box<dyn Error>> {
            chart.populate_variables(config.variables.clone());
            // First step is to pull the chart to the working dir
            let current_repo = chart.chart.find_repo(config.repositories.clone())?;
            let chart_path = current_repo.pull_chart(chart.chart.clone(), workdir_path.clone())?;
            if let Some(extensions) = chart.extensions.clone() {
                extensions
                    .into_iter()
                    .try_for_each(|extension| -> Result<(), Box<dyn Error>> {
                        extension.apply(chart_path.clone(), config_full_path.clone())
                    })?
            }
            if let Some(patches) = chart.patches.clone() {
                patches
                    .into_iter()
                    .try_for_each(|patch| -> Result<(), Box<dyn Error>> {
                        patch.apply(chart_path.clone(), config.patches.clone(), workdir_path.clone(), args.git_bin.clone())
                    })?
            }
            config
                .mirrors
                .iter()
                .try_for_each(|mirror| -> Result<(), Box<dyn Error>> {
                    mirror_from_mirror_obj(mirror.clone())?.push(
                        workdir_path.clone(),
                        chart_path.clone(),
                        chart.clone(),
                        args.dry_run,
                    )
                })?;
            message_empty(&format!("{}", chart.chart.name));
            Ok(())
        })?;
    todo!()
}

fn main() {
    match exec(Args::parse()) {
        Ok(()) => message_empty("Thanks for using helmule"),
        Err(err) => {
            message_error(err);
            exit(1)
        }
    }
}
