use std::{
    error::Error,
    fs::{self, read_dir},
    path::{Path, PathBuf},
    process::Command,
};

use which::which;

use crate::output::message_info;

pub fn cli_exec(command: String) -> Result<String, Box<dyn Error>> {
    message_info(&format!("executing: {}", command));
    let expect = format!("command has failed: {}", command);
    let output = Command::new("sh")
        .arg("-c")
        .arg(command)
        .output()
        .expect(&expect);
    let stderr = String::from_utf8_lossy(&output.stderr);
    if !&output.status.success() {
        return Err(Box::from(stderr));
    };
    Ok(String::from_utf8_lossy(&output.stdout).to_string())
}

pub fn cli_exec_from_dir(command: String, dir: String) -> Result<String, Box<dyn Error>> {
    message_info(&format!("executing: {} from {}", command, dir));
    let expect = format!("command has failed: {}", command);
    let output = Command::new("sh")
        .arg("-c")
        .current_dir(dir)
        .arg(command)
        .output()
        .expect(&expect);
    let stderr = String::from_utf8_lossy(&output.stderr);
    if !&output.status.success() {
        return Err(Box::from(stderr));
    };
    let mut stdout = String::from_utf8_lossy(&output.stdout).to_string();
    stdout.pop();
    Ok(stdout)
}

// A helper that checks wheter all the required binaries are installed
pub fn check_prerequisites(bins: Vec<String>) -> Result<(), Box<dyn Error>> {
    message_info(&"checking prerequisites".to_string());
    for bin in bins {
        message_info(&format!("checking {}", bin));
        which(bin)?;
    }
    Ok(())
}

pub fn get_full_path(rel_path: String) -> Result<String, Box<dyn Error>> {
    match PathBuf::from(&rel_path)
        .canonicalize()?
        .into_os_string()
        .into_string()
    {
        Ok(path) => Ok(path),
        Err(_) => Err(Box::from(format!(
            "{} can't be converted into a full path",
            rel_path
        ))),
    }
}

pub fn get_full_path_dir(rel_path: String) -> Result<String, Box<dyn Error>> {
    let res = match PathBuf::from(&rel_path).parent() {
        Some(res) => res.canonicalize()?.into_os_string().into_string(),
        None => PathBuf::from(&rel_path)
            .canonicalize()?
            .into_os_string()
            .into_string(),
    };
    match res {
        Ok(path) => Ok(path),
        Err(_) => Err(Box::from(format!(
            "{} can't be converted into a full path",
            rel_path
        ))),
    }
}

/// Copy files from source to destination recursively.
pub fn copy_recursively(
    source: impl AsRef<Path>,
    destination: impl AsRef<Path>,
) -> Result<(), Box<dyn Error>> {
    for entry in read_dir(source)? {
        let entry = entry?;
        let filetype = entry.file_type()?;
        if filetype.is_dir() {
            copy_recursively(entry.path(), destination.as_ref().join(entry.file_name()))?;
        } else {
            message_info(&format!("trying to copy {:?}", entry.path()));
            fs::copy(entry.path(), destination.as_ref().join(entry.file_name()))?;
        }
    }
    Ok(())
}

pub fn is_path_relative(path: String) -> bool {
    !path.starts_with('/')
}

#[cfg(test)]
mod tests {
    use super::{cli_exec, cli_exec_from_dir};
    use tempfile::TempDir;

    #[test]
    fn test_stderr() {
        let command = ">&2 echo \"error\" && exit 1";
        let test = cli_exec(command.to_string());
        assert_eq!(test.err().unwrap().to_string(), "error\n".to_string());
    }

    #[test]
    fn test_stdout() {
        let command = "echo test";
        let test = cli_exec(command.to_string());
        assert_eq!(test.unwrap().to_string(), "test\n".to_string());
    }

    #[test]
    fn test_stdout_current_dir() {
        let dir = TempDir::new().unwrap();
        let dir_str = dir.path().to_str().unwrap().to_string();
        let command = "echo $PWD";
        let test = cli_exec_from_dir(command.to_string(), dir_str.clone());
        assert!(test.unwrap().to_string().contains(dir_str.as_str()));
    }

    #[test]
    fn test_stderr_current_dir() {
        let dir = TempDir::new().unwrap();
        let dir_str = dir.path().to_str().unwrap().to_string();
        let command = ">&2 echo \"error\" && exit 1";
        let test = cli_exec_from_dir(command.to_string(), dir_str.clone());
        assert_eq!(test.err().unwrap().to_string(), "error\n".to_string());
    }
}
