use std::{error::Error, ffi::OsStr, fs::File, path::Path};

use serde::de::DeserializeOwned;

pub trait ConfigImpl {
    fn apply_includes(&mut self, config_path: String) -> Result<(), Box<dyn Error>>;
}

pub fn read_config<T: DeserializeOwned>(path: String) -> Result<T, Box<dyn Error>> {
    let config_content = File::open(path.clone())?;
    let config = match get_extension_from_filename(&path) {
        Some(ext) => match ext {
            "yaml" | "yml" => serde_yaml::from_reader(config_content)?,
            _ => return Err(Box::from(format!("{} files are not supported", ext))),
        },
        None => return Err(Box::from("can't read file without extension")),
    };
    Ok(config)
}

fn get_extension_from_filename(filename: &str) -> Option<&str> {
    Path::new(filename).extension().and_then(OsStr::to_str)
}

#[cfg(test)]
mod tests {
    use super::{get_extension_from_filename, read_config};
    use serde::{Deserialize, Serialize};
    use std::{error::Error, fs::File, io::Write};
    use tempfile::tempdir;

    #[test]
    fn test_extension_getter() {
        let filepath = "/tmp/config.yaml";
        let extension = get_extension_from_filename(filepath);
        assert_eq!(extension, Some("yaml"));
    }

    #[test]
    fn test_extension_getter_empty() {
        let filepath = "/tmp/config";
        let extension = get_extension_from_filename(filepath);
        assert_eq!(extension, None);
    }

    #[derive(Serialize, Deserialize, PartialEq, Debug, Clone)]
    struct DummyConfig {
        string: String,
        amounts: Vec<DummyProperty>,
    }

    #[derive(Serialize, Deserialize, PartialEq, Debug, Clone)]
    struct DummyProperty {
        amount: i32,
    }

    fn prepare_test_file(name: &str, data: &str) -> Result<String, Box<dyn Error>> {
        let dir = tempdir()?;
        let file_path = dir.into_path().join(&name);
        let mut file = File::create(file_path.clone())?;
        file.write_all(data.as_bytes())?;
        let path = file_path.into_os_string().to_str().unwrap().to_string();
        Ok(path)
    }

    #[test]
    fn test_config_reader() -> Result<(), Box<dyn Error>> {
        let content = "---
            string: test
            amounts: 
              - amount: 4
              - amount: 5 
            ";
        let file_path = prepare_test_file("config.yaml", content)?;
        let config_data: DummyConfig;
        config_data = read_config(file_path)?;

        let expected = DummyConfig {
            string: "test".to_string(),
            amounts: vec![DummyProperty { amount: 4 }, DummyProperty { amount: 5 }],
        };

        assert_eq!(expected, config_data);
        Ok(())
    }
}
