use std::{error::Error, fs::rename};

use base64::{engine::general_purpose, Engine};
use serde::{Deserialize, Serialize};

use crate::cli::{cli_exec, cli_exec_from_dir};

use super::{
    chart::Chart,
    repository::{RepositoryImpl, Version},
};

// A struct that represents a regular helm repo
#[derive(Serialize, Deserialize, PartialEq, Debug, Clone)]
pub struct HelmRepo {
    // A URL of the helm repository
    pub url: String,
}

const LATEST_VERSION: &str = "latest";

impl RepositoryImpl for HelmRepo {
    fn pull_chart(&self, chart: Chart, workdir_path: String) -> Result<String, Box<dyn Error>> {
        match self.repo_kind_from_url(self.clone())? {
            RepoKind::Default => self.pull_default(chart, workdir_path),
            RepoKind::Oci => self.pull_oci(chart, workdir_path),
        }
    }

    fn get_url(&self) -> String {
        self.url.clone()
    }
}

pub(crate) enum RepoKind {
    Default,
    Oci,
}

impl HelmRepo {
    fn repo_kind_from_url(&self, repository: HelmRepo) -> Result<RepoKind, Box<dyn Error>> {
        let prefix = repository
            .url
            .chars()
            .take_while(|&ch| ch != ':')
            .collect::<String>();
        match prefix.as_str() {
            "oci" => Ok(RepoKind::Oci),
            "https" | "http" => Ok(RepoKind::Default),
            _ => Err(Box::from(format!(
                "repo kind is not defined by the prefix: {}",
                prefix
            ))),
        }
    }

    fn pull_oci(&self, chart: Chart, workdir_path: String) -> Result<String, Box<dyn Error>> {
        let args = match chart.version.as_str() {
            LATEST_VERSION => "".to_string(),
            _ => format!("--version {}", chart.version.clone()),
        };
        let repo = match self.get_url().ends_with('/') {
            true => {
                let mut repo = self.get_url().clone();
                repo.pop();
                repo
            }
            false => self.get_url().clone(),
        };
        let cmd = format!(
            "helm pull {}/{} {} --destination {} --untar",
            repo, chart.name, args, workdir_path
        );
        cli_exec(cmd)?;
        // Get the version
        let cmd = format!("helm show chart {}/{}", workdir_path, chart.name);
        let helm_stdout = cli_exec(cmd)?;
        let old_dir_name = format!("{}/{}", workdir_path, chart.name);
        let new_dir_name: String;
        match serde_yaml::from_str::<Version>(&helm_stdout) {
            Ok(res) => {
                new_dir_name = format!("{}-{}", old_dir_name, res.version);
                rename(old_dir_name, new_dir_name.clone())?;
            }
            Err(err) => return Err(Box::from(err)),
        };

        // TODO: Do we really need it?
        let cmd = "helm show chart . | yq '.version'".to_string();
        let _version = cli_exec_from_dir(cmd, new_dir_name.clone())?;
        Ok(new_dir_name)
    }

    fn pull_default(&self, chart: Chart, workdir_path: String) -> Result<String, Box<dyn Error>> {
        // Add repo and update
        let repo_local_name = general_purpose::STANDARD_NO_PAD.encode(self.get_url());
        let cmd = format!("helm repo add {} {}", repo_local_name, self.get_url());
        cli_exec(cmd)?;
        cli_exec("helm repo update".to_string())?;

        let args = match chart.version.as_str() {
            LATEST_VERSION => "".to_string(),
            _ => format!("--version {}", chart.version.clone()),
        };
        let cmd = format!(
            "helm pull {}/{} {} --destination {} --untar",
            repo_local_name, chart.name, args, workdir_path
        );
        cli_exec(cmd)?;

        // Get the version
        let cmd = format!("helm show chart {}/{}", workdir_path, chart.name);
        let helm_stdout = cli_exec(cmd)?;
        let old_dir_name = format!("{}/{}", workdir_path, chart.name);
        let new_dir_name: String;
        match serde_yaml::from_str::<Version>(&helm_stdout) {
            Ok(res) => {
                new_dir_name = format!("{}-{}", old_dir_name, res.version);
                rename(old_dir_name, new_dir_name.clone())?;
            }
            Err(err) => return Err(Box::from(err)),
        };

        //cleaning up
        let cmd = format!("helm repo remove {}", repo_local_name);
        cli_exec(cmd)?;

        // TODO: Do we really need it?
        let cmd = "helm show chart . | yq '.version'".to_string();
        let _version = cli_exec_from_dir(cmd, new_dir_name.clone())?;
        Ok(new_dir_name)
    }
}
